/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/

#include "RveFlatModel.h"
#include "RveCartesianModel.h"
#include "RveIsoModel.h"
#include "GLWidget.h"

#include <gl/GLU.h>

#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE  0x809D
#endif

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
GLWidget::GLWidget( QWidget *parent ) :
   QGLWidget( QGLFormat( QGL::SampleBuffers ), parent ),
   m_xRot( 0.0 ),
   m_yRot( 0.0 ),
   m_zRot( 0.0 ),
   m_Scale( 1.0 ),
   m_Dx( 0.0 ),
   m_Dy( 0.0 ),
   m_LeftButtonPressed( false ),
   m_pRawdata( 0 ),
   m_pFlatModel( 0 ),
   m_pCartesianModel( 0 ),
   m_pIsoModel( 0 ),
   m_pMap( 0 ),
   m_MSecs( -1 ),
   m_MouseMoveMSecs( -1 )
{
   m_pMap = new RveMap( this );
   m_pTimer = new QTimer( this );
   m_pTimer->setInterval( 0 );
   connect( m_pTimer, SIGNAL( timeout() ), this, SLOT( render() ) );
   m_pTimer->start();
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
GLWidget::~GLWidget()
{
   clearModels();
   if( m_pMap )
      delete m_pMap;
   if( m_pTimer )
      delete m_pTimer;
}


void GLWidget::render()
{
   bool ok = m_MSecs > 0;


   quint64 curMSecs = QDateTime::currentMSecsSinceEpoch();
   quint64 prevMSecs = m_MSecs;

   m_MSecs = curMSecs;

   if( !ok )
      return;

   quint64 msecsElapsed = curMSecs - prevMSecs;
   double secsElapsed = (double)msecsElapsed / 1000.0;

   if( !m_LeftButtonPressed )
   {
      m_xRot += secsElapsed * m_Dy;
      m_zRot += secsElapsed * m_Dx;
      double d = pow( 0.25, secsElapsed );

      if( fabs( m_Dy ) < 2.0 )
      {
         m_Dy = 0.0;
      } else
      {
         m_Dy *= d;
      }

      if( fabs( m_Dx ) < 2.0 )
      {
         m_Dx = 0.0;
      } else
      {
         m_Dx *= d;
      }

   }

   updateGL();
}


RveFlatModel *GLWidget::flatModel()
{
   if( !m_pFlatModel )
      m_pFlatModel = new RveFlatModel();

   return( m_pFlatModel );
}


void GLWidget::enableFlatModel()
{
   flatModel()->enable();
   cartesianModel()->disable();
   isoModel()->disable();
}


bool GLWidget::flatModelEnabled()
{
   return( flatModel()->isEnabled() );
}


void GLWidget::setFlatParameters( RveFlatParameters parameters )
{
   flatModel()->setParameters( parameters );
   updateGL();
}


RveFlatParameters GLWidget::flatParameters()
{
   return( flatModel()->parameters() );
}


RveCartesianModel *GLWidget::cartesianModel()
{
   if( !m_pCartesianModel )
      m_pCartesianModel = new RveCartesianModel( false );

   return( m_pCartesianModel );
}


void GLWidget::enableCartesianModel()
{
   isoModel()->disable();
   cartesianModel()->enable();
   flatModel()->disable();
}


bool GLWidget::cartesianModelEnabled()
{
   return( cartesianModel()->isEnabled() );
}


void GLWidget::setCartesianParameters( RveCartesianParameters parameters )
{
   cartesianModel()->setParameters( parameters );
   updateGL();
}


RveCartesianParameters GLWidget::cartesianParameters()
{
   return( cartesianModel()->parameters() );
}


RveIsoModel *GLWidget::isoModel()
{
   if( !m_pIsoModel )
      m_pIsoModel = new RveIsoModel();

   return( m_pIsoModel );
}


void GLWidget::enableIsoModel()
{
   isoModel()->enable();
   cartesianModel()->disable();
   flatModel()->disable();
}


bool GLWidget::isoModelEnabled()
{
   return( isoModel()->isEnabled() );
}


void GLWidget::setIsoParameters( RveIsoParameters parameters )
{
   isoModel()->setParameters( parameters );
   updateGL();
}


RveIsoParameters GLWidget::isoParameters()
{
   return( isoModel()->parameters() );
}


void GLWidget::onNewRawdata( RveRawdata *pRawdata )
{
   m_xRot = 0.0;
   m_yRot = 0.0;
   m_zRot = 0.0;
   m_Scale = 1.0;
   m_Dx = 0.0;
   m_Dy = 0.0;

   setData( pRawdata );
}

void GLWidget::setData( RveRawdata *pRawdata )
{
   if( m_pRawdata == pRawdata )
      return;

   m_pRawdata = pRawdata;

   flatModel()->setRawdata( pRawdata );
   cartesianModel()->setRawdata( pRawdata );
   isoModel()->setRawdata( pRawdata );

   m_pMap->setLonLat( pRawdata->lon(), pRawdata->lat() );
}


RveMap *GLWidget::map()
{
   return( m_pMap );
}


void GLWidget::clearModels()
{
   if( m_pFlatModel )
   {
      delete m_pFlatModel;
      m_pFlatModel = 0;
   }

   if( m_pIsoModel )
   {
      delete m_pIsoModel;
      m_pIsoModel = 0;
   }

   if( m_pCartesianModel )
   {
      delete m_pCartesianModel;
      m_pCartesianModel = 0;
   }
}


void GLWidget::initializeGL()
{
   initializeGLFunctions();

   glClearColor( 0.0, 0.0, 0.0, 0.0 );

   glEnable( GL_BLEND );
   glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
   glEnable( GL_DEPTH_TEST );
   glDisable( GL_CULL_FACE );
   glShadeModel( GL_FLAT );
}


void GLWidget::resizeGL( int width, int height )
{
   int side = qMin( width, height );
   glViewport( 0, 0, width, height );

   glMatrixMode( GL_PROJECTION );
   glLoadIdentity();

   double aspect = (double)width / (double)height;
   gluPerspective( 70.0, aspect, 1.5, 200000.0 );
   glMatrixMode( GL_MODELVIEW );
}


void GLWidget::paintGL()
{
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
   glLoadIdentity();

/*   static GLfloat lightSpecular[4] = { 0.0, 0.0, 0.0, 1.0 };
   glLightfv( GL_LIGHT0, GL_SPECULAR, lightSpecular );*/
   static GLfloat lightDiffuse[4] = { 0.004, 0.004, 0.004, 0.0 };
   glLightfv( GL_LIGHT0, GL_DIFFUSE, lightDiffuse );
/*   static GLfloat lightAmbient[4] = { 0.0, 0.0, 0.0, 1.0 };
   glLightfv( GL_LIGHT0, GL_AMBIENT, lightAmbient );*/
   static GLfloat lightPosition[4] = { 0.0, 0.0, 0.0, 1.0 };
   glLightfv( GL_LIGHT0, GL_POSITION, lightPosition );

   glPushMatrix();

   glTranslatef( 0.0, 0.0, -150.0 );
   glRotatef( m_xRot, 1.0, 0.0, 0.0 );
   glRotatef( m_yRot, 0.0, 1.0, 0.0 );
   glRotatef( m_zRot, 0.0, 0.0, 1.0 );
   glScalef( 0.001 * m_Scale, 0.001 * m_Scale, 0.0075 * m_Scale );

   if( flatModel()->isEnabled() )
   {
      glDisable( GL_LIGHTING );
      glDisable( GL_LIGHT0 );
      glDisable( GL_CULL_FACE );
      flatModel()->paintGL();
   }

   if( cartesianModel()->isEnabled() )
   {
      glDisable( GL_LIGHTING );
      glDisable( GL_LIGHT0 );
      glDisable( GL_CULL_FACE );
      cartesianModel()->paintGL();
   }

   if( isoModel()->isEnabled() )
   {
      glEnable( GL_BLEND );
      glEnable( GL_ALPHA );
      isoModel()->paintGL();
   }

   if( m_pMap )
   {
      m_pMap->paintGL();
   }

   // Paint 2D Elements
   if( true )
   {
/*      glPointSize( 10.0 );
      glEnable( GL_POINT_SMOOTH );
      glColor3f( (float) 0.9, (float) 0.9, (float) 0.9 );
      glBegin( GL_POINTS );
      for( int i = 0; i < m_Cities.size(); i++ )
      {
         RbGLSprite *pCitySprite = m_Cities[i];
         glVertex3f( pCitySprite->getX(), pCitySprite->getY(), pCitySprite->getZ() );
      }
      glEnd();
      glColor3f( 1.0, 1.0, 1.0 );

      for( int i = 0; i < m_Cities.size(); i++ )
      {
         m_Cities[i]->projectTo2D();
      }

      for( int i = 0; i < m_Sprites.size(); i++ )
      {
         m_Sprites[i]->projectTo2D();
      }

      enter2DMode();
   
      for( int i = 0; i < m_Sprites.size(); i++ )
      {
         m_Sprites[i]->paint();
      }

      for( int i = 0; i < m_Cities.size(); i++ )
      {
         m_Cities[i]->paint();
      }

      leave2DMode();*/
   }

   glPopMatrix();
}


void GLWidget::mousePressEvent( QMouseEvent *event )
{
   if( event->button() & Qt::LeftButton )
   {
      m_LeftButtonPressed = true;
      m_MouseMoveMSecs = QDateTime::currentMSecsSinceEpoch();

      m_MouseX = event->x();
      m_MouseY = event->y();
   }
}

void GLWidget::mouseReleaseEvent( QMouseEvent *event )
{
   if( event->button() & Qt::LeftButton )
   {
      m_LeftButtonPressed = false;
      qint64 curMSecs = QDateTime::currentMSecsSinceEpoch();
      qint64 msecsElapsed = curMSecs - m_MouseMoveMSecs;
      double secsElapsed = (double)msecsElapsed / 1000.0;
      m_MouseMoveMSecs = -1;

      if( secsElapsed > 0.1 )
      {
         m_Dx = 0.0;
         m_Dy = 0.0;
      }
   }
}

void GLWidget::mouseMoveEvent( QMouseEvent *event )
{
   double dx = (double)(event->x() - m_MouseX) / (double)width();
   double dy = (double)(event->y() - m_MouseY) / (double)height();
   qint64 curMSecs = QDateTime::currentMSecsSinceEpoch();
   qint64 msecsElapsed = curMSecs - m_MouseMoveMSecs;
   double secsElapsed = (double)msecsElapsed / 1000.0;

   m_Dx = 200.0 * dx;
   m_Dy = 200.0 * dy;
   m_xRot += m_Dy;
   m_zRot += m_Dx;

   m_Dx /= secsElapsed;
   m_Dy /= secsElapsed;

   updateGL();

   m_MouseX = event->x();
   m_MouseY = event->y();

   m_MouseMoveMSecs = curMSecs;
}


void GLWidget::wheelEvent( QWheelEvent *event )
{
   m_Scale += (double)event->delta() / 10000.0;

   updateGL();
}


void GLWidget::project3DTo2D(
   double xPos, double yPos, double zPos,
   double &resX, double &resY, double &resZ )
{
   GLdouble projection[16];
   GLdouble modelview[16];
   GLdouble screencoords[3];
   GLint viewport[4];
   glGetDoublev( GL_PROJECTION_MATRIX, projection );
   glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
   glGetIntegerv( GL_VIEWPORT, viewport );
   gluProject( xPos, yPos, zPos,
               modelview, projection,
               viewport,
               &screencoords[0], &screencoords[1], &screencoords[2] );
               
   resX = screencoords[0];
   resY = screencoords[1];
   resZ = screencoords[2];

   double clipZ = ( 2.0 * resZ ) - 1.0;
   double zFar = 1000.0;
   double zNear = 0.0;
   resZ = ( ( clipZ + ( zFar + zNear ) / ( zFar - zNear ) ) * ( zFar - zNear ) ) / ( -2.0 );
}


void GLWidget::enter2DMode()
{
   glMatrixMode( GL_PROJECTION );
   glPushMatrix();
   glLoadIdentity();
   glOrtho( 0.0, width(), 0.0, height(), 0.0, 1000.0 );
   glViewport( 0, 0, width(), height() );
   glMatrixMode( GL_MODELVIEW );
   glPushMatrix();
   glLoadIdentity();

   glEnable( GL_TEXTURE_2D );
   glEnable( GL_BLEND );
   glEnable( GL_ALPHA );
   glEnable( GL_DEPTH_TEST );
   glDepthMask( GL_FALSE );
}


void GLWidget::leave2DMode()
{
   glDepthMask( GL_TRUE );
   glDisable( GL_TEXTURE_2D );
   glDisable( GL_BLEND );
   glDisable( GL_ALPHA );

   glPopMatrix();

   glMatrixMode( GL_PROJECTION );
   glPopMatrix();
   glMatrixMode( GL_MODELVIEW );
}
