/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/

#include <QVBoxLayout>
#include <QSizePolicy>

#include <QFileDialog>
#include <QCheckBox>
#include <QMessageBox>
#include <QComboBox>

#include "MainWindow.h"
#include "RveAboutDialog.h"
#include "RveRawdataSettings.h"

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
MainWindow::MainWindow() :
   m_pRawdata( 0 )
{
   setupUi( this );
   m_pGLWidget = new GLWidget( this );
   m_pGLWidget->setSizePolicy( QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding ) );
   glLayout->addWidget( m_pGLWidget );
   connect( this, SIGNAL( newRawdata( RveRawdata * ) ), m_pGLWidget, SLOT( onNewRawdata( RveRawdata * ) ) );

   splitter->setStretchFactor( 0, 1 );
   splitter->setStretchFactor( 1, 0 );

   twIsoParameters->header()->setResizeMode( QHeaderView::ResizeToContents );

   m_pLayersGui = new RveLayersGuiHandler( twLayerParameters, m_pGLWidget->map(), this );
   connect( m_pLayersGui, SIGNAL( updateGL() ), m_pGLWidget, SLOT( updateGL() ) );
   connect( this, SIGNAL( newRawdata( RveRawdata * ) ), m_pLayersGui, SLOT( onNewRawdata( RveRawdata * ) ) );

   updateGui();

   connect( actionOpen, SIGNAL( triggered() ), this, SLOT( onFileOpen() ) );
   connect( actionAbout, SIGNAL( triggered() ), this, SLOT( onAbout() ) );

   // Flat Model
   connect( sbFlatClipMin, SIGNAL( valueChanged( double ) ), this, SLOT( sbClipMinMaxFlatModelUpdated() ) );
   connect( sbFlatClipMax, SIGNAL( valueChanged( double ) ), this, SLOT( sbClipMinMaxFlatModelUpdated() ) );
   connect( pbFlatApply, SIGNAL( clicked() ), this, SLOT( onPbFlatApply() ) );

   // Cartesian Model
   connect( sbCartesianClipMin, SIGNAL( valueChanged( double ) ), this, SLOT( onCartesianUpdated() ) );
   connect( sbCartesianClipMax, SIGNAL( valueChanged( double ) ), this, SLOT( onCartesianUpdated() ) );
   connect( sbCartesianNCells, SIGNAL( valueChanged( int ) ), this, SLOT( onCartesianUpdated() ) );
   connect( sbCartesianNLayers, SIGNAL( valueChanged( int ) ), this, SLOT( onCartesianUpdated() ) );
   connect( sbCartesianLowestLayer, SIGNAL( valueChanged( double ) ), this, SLOT( onCartesianUpdated() ) );
   connect( sbCartesianHighestLayer, SIGNAL( valueChanged( double ) ), this, SLOT( onCartesianUpdated() ) );
   connect( pbCartesianApply, SIGNAL( clicked() ), this, SLOT( onPbCartesianApply() ) );

   // Iso Model
   connect( pbIsoAdd, SIGNAL( clicked() ), this, SLOT( onPbIsoAdd() ) );
   connect( pbIsoApply, SIGNAL( clicked() ), this, SLOT( onPbIsoApply() ) );
   connect( sbIsoNCells, SIGNAL( valueChanged( int ) ), this, SLOT( onIsoUpdated() ) );
   connect( sbIsoNLayers, SIGNAL( valueChanged( int ) ), this, SLOT( onIsoUpdated() ) );
   connect( sbIsoLowestLayer, SIGNAL( valueChanged( double ) ), this, SLOT( onIsoUpdated() ) );
   connect( sbIsoHighestLayer, SIGNAL( valueChanged( double ) ), this, SLOT( onIsoUpdated() ) );

   connect( toolBox, SIGNAL( currentChanged( int ) ), this, SLOT( onChangeModelType() ) );
   connect( toolBox, SIGNAL( currentChanged( int ) ), this, SLOT( onChangeModelType() ) );
   connect( toolBox, SIGNAL( currentChanged( int ) ), this, SLOT( onChangeModelType() ) );
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
MainWindow::~MainWindow()
{
   delete m_pLayersGui;
}


void MainWindow::onChangeModelType()
{
   updateGui();
}


void MainWindow::onAbout()
{
   RveAboutDialog aboutDialog( this );
   aboutDialog.setModal( true );
   aboutDialog.show();
   aboutDialog.exec();
}

void MainWindow::onFileOpen()
{
   QString fileName = QFileDialog::getOpenFileName(
      this,
      "Please Select a Raw Data File",
      QCoreApplication::applicationDirPath() + QDir::separator() + "..",
      QString( "HDF5 Volume Files (*.h5)" ) );

   if( !fileName.isEmpty() )
   {
      RveRawdata *pRawdata = new RveRawdata( fileName, this );
      if( pRawdata->isValid() )
      {
         QString dataMoment;
         if( pRawdata->availableDataMoments().size() == 1 )
         {
            dataMoment = pRawdata->availableDataMoments()[0];
         } else
         {
            RveRawdataSettings rds( pRawdata, this );
            rds.setModal( true );
            rds.show();
            rds.exec();
            dataMoment = rds.selectedMoment();
         }

         if( m_pRawdata )
         {
            delete m_pRawdata;
         }
         m_pRawdata = pRawdata;

         m_pRawdata->loadVolume( dataMoment );
         info->setText( m_pRawdata->text() );

         double min = m_pRawdata->minValue();
         double max = m_pRawdata->maxValue();

         bool old = sbFlatClipMin->blockSignals( true );
         sbFlatClipMin->setValue( min );
         sbFlatClipMin->blockSignals( old );

         old = sbFlatClipMax->blockSignals( true );
         sbFlatClipMax->setValue( max );
         sbFlatClipMax->blockSignals( old );

         old = sbCartesianClipMin->blockSignals( true );
         sbCartesianClipMin->setValue( min );
         sbCartesianClipMin->blockSignals( old );

         old = sbCartesianClipMax->blockSignals( true );
         sbCartesianClipMax->setValue( max );
         sbCartesianClipMax->blockSignals( false );

         emit newRawdata( m_pRawdata );

         updateGui();
      } else
      {
         delete pRawdata;
         QMessageBox::critical( this, "File Load", QString( "Couldn't load File\n" ) + fileName );
      }
   }
}


void MainWindow::updateGui()
{
//   wFlatModel->setVisible( flatModelEnabled() );
//   wCartesianModel->setVisible( cartesianModelEnabled() );

   m_pLayersGui->updateGui();

   if( m_pRawdata )
   {
      toolBox->setEnabled( true );

      // Enable/Disable Flat Model GUI
      lFlatClipMin->setEnabled( flatModelEnabled() );
      lFlatClipMax->setEnabled( flatModelEnabled() );
      sbFlatClipMin->setEnabled( flatModelEnabled() );
      sbFlatClipMax->setEnabled( flatModelEnabled() );

      // Enable/Disable Cartesian Model GUI
      lCartesianClipMin->setEnabled( cartesianModelEnabled() );
      sbCartesianClipMin->setEnabled( cartesianModelEnabled() );
      lCartesianClipMax->setEnabled( cartesianModelEnabled() );
      sbCartesianClipMax->setEnabled( cartesianModelEnabled() );
      lCartesianNCells->setEnabled( cartesianModelEnabled() );
      sbCartesianNCells->setEnabled( cartesianModelEnabled() );
      lCartesianNLayers->setEnabled( cartesianModelEnabled() );
      sbCartesianNLayers->setEnabled( cartesianModelEnabled() );
      lCartesianHighestLayer->setEnabled( cartesianModelEnabled() );
      sbCartesianHighestLayer->setEnabled( cartesianModelEnabled() );
      lCartesianLowestLayer->setEnabled( cartesianModelEnabled() );
      sbCartesianLowestLayer->setEnabled( cartesianModelEnabled() );

      // Enable/Disable Iso Model GUI
      lIsoNCells->setEnabled( isoModelEnabled() );
      sbIsoNCells->setEnabled( isoModelEnabled() );
      lIsoNLayers->setEnabled( isoModelEnabled() );
      sbIsoNLayers->setEnabled( isoModelEnabled() );
      lIsoLowestLayer->setEnabled( isoModelEnabled() );
      sbIsoLowestLayer->setEnabled( isoModelEnabled() );
      lIsoHighestLayer->setEnabled( isoModelEnabled() );
      sbIsoHighestLayer->setEnabled( isoModelEnabled() );
      twIsoParameters->setEnabled( isoModelEnabled() );
      pbIsoAdd->setEnabled( isoModelEnabled() );
      pbIsoApply->setEnabled( isoModelEnabled() );

      if( !flatModelEnabled() )
         pbFlatApply->setEnabled( false );
      if( !cartesianModelEnabled() )
         pbCartesianApply->setEnabled( false );
      if( !isoModelEnabled() )
         pbIsoApply->setEnabled( false );

      if( flatModelEnabled() )
      {
         m_pGLWidget->setFlatParameters( flatParameters() );
         if( !m_pGLWidget->flatModelEnabled() )
         {
            m_pGLWidget->enableFlatModel();
         }

         pbFlatApply->setEnabled(
            m_pGLWidget->flatParameters() != flatParameters() );
      } else
      if( cartesianModelEnabled() )
      {
         m_pGLWidget->setCartesianParameters( cartesianParameters() );
         if( !m_pGLWidget->cartesianModelEnabled() )
         {
            m_pGLWidget->enableCartesianModel();
         }

         pbCartesianApply->setEnabled(
            m_pGLWidget->cartesianModel()->parameters() != cartesianParameters() );
      } else
      if( isoModelEnabled() )
      {
         m_pGLWidget->setIsoParameters( isoParameters() );
         if( !m_pGLWidget->isoModelEnabled() )
         {
            m_pGLWidget->enableIsoModel();
         }

         pbIsoApply->setEnabled(
            m_pGLWidget->isoModel()->parameters() != isoParameters() );
      }
   } else
   {
      toolBox->setEnabled( false );

      lFlatClipMin->setEnabled( false );
      lFlatClipMax->setEnabled( false );
      sbFlatClipMin->setEnabled( false );
      sbFlatClipMax->setEnabled( false );
   }

   m_pGLWidget->updateGL();
}


GLWidget *MainWindow::glWidget()
{
   return( m_pGLWidget );
}


bool MainWindow::flatModelEnabled() const
{
   return( toolBox->currentIndex() == 0  );
}


bool MainWindow::cartesianModelEnabled() const
{
   return( toolBox->currentIndex() == 1 );
}


bool MainWindow::isoModelEnabled() const
{
   return( toolBox->currentIndex() == 2 );
}


void MainWindow::sbClipMinMaxFlatModelUpdated()
{
   pbFlatApply->setEnabled( m_pGLWidget->flatParameters() != flatParameters() );
}


void MainWindow::onCartesianUpdated()
{
   pbCartesianApply->setEnabled( m_pGLWidget->cartesianParameters() != cartesianParameters() );
}

void MainWindow::onPbFlatApply()
{
   m_pGLWidget->setFlatParameters( flatParameters() );
   updateGui();
}


void MainWindow::onPbCartesianApply()
{
   m_pGLWidget->setCartesianParameters( cartesianParameters() );
   updateGui();
}


void MainWindow::onPbIsoApply()
{
   m_pGLWidget->setIsoParameters( isoParameters() );
   updateGui();
}


RveFlatParameters MainWindow::flatParameters() const
{
   return( RveFlatParameters( sbFlatClipMin->value(), sbFlatClipMax->value() ) );
}


RveCartesianParameters MainWindow::cartesianParameters() const
{
   return( RveCartesianParameters(
      sbCartesianNCells->value(),
      sbCartesianNLayers->value(),
      sbCartesianLowestLayer->value() * 1000.0,
      sbCartesianHighestLayer->value() * 1000.0,
      sbCartesianClipMin->value(),
      sbCartesianClipMax->value() ) );

}


RveIsoParameters MainWindow::isoParameters() const
{
   RveIsoParameters p(
      sbIsoNCells->value(),
      sbIsoNLayers->value(),
      sbIsoLowestLayer->value() * 1000.0,
      sbIsoHighestLayer->value() * 1000.0 );

   foreach( QTreeWidgetItem *twi, m_IsoParameterEntries.keys() )
   {
      RveIsoParameterEntry *pEntry = m_IsoParameterEntries.value( twi );
      if( pEntry->isEnabled() )
      {
         p.addIsoLevel( pEntry->isoParameters() );
      }
   }

   return( p );
}


void MainWindow::onPbIsoAdd()
{
   RveIsoParameterEntry *isoParameterEntry = new RveIsoParameterEntry( twIsoParameters);
   connect( isoParameterEntry, SIGNAL( remove(  RveIsoParameterEntry * ) ), this, SLOT( onPbIsoRemove( RveIsoParameterEntry * ) ) );
   connect( isoParameterEntry, SIGNAL( changed( RveIsoParameterEntry * ) ), this, SLOT( onIsoChanged( RveIsoParameterEntry * ) ) );
   m_IsoParameterEntries.insert( isoParameterEntry->treeWidgetItem(), isoParameterEntry );

   onIsoUpdated();
}

void MainWindow::onPbIsoRemove( RveIsoParameterEntry *isoParameterEntry )
{
   m_IsoParameterEntries.remove( isoParameterEntry->treeWidgetItem() );
   delete isoParameterEntry->treeWidgetItem();
   delete isoParameterEntry;

   onIsoUpdated();
}


void MainWindow::onIsoChanged( RveIsoParameterEntry *isoParameterEntry )
{
   onIsoUpdated();
}


void MainWindow::onTwIsoColor( QPushButton *color )
{
}


void MainWindow::onIsoUpdated()
{
   pbIsoApply->setEnabled( m_pGLWidget->isoParameters() != isoParameters() );
}
