/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/

#include <QImage>
#define _USE_MATH_DEFINES
#include <math.h>

#include "RveCartesianModel.h"

#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE  0x809D
#endif

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveCartesianModel::RveCartesianModel( bool keepData ) :
   RveModel( 0 ),
   m_Parameters( RveCartesianParameters() ),
   m_KeepData( keepData ),
   m_Enabled( false )
{
   initializeGLFunctions();
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveCartesianModel::~RveCartesianModel()
{
}


void RveCartesianModel::enable()
{
   if( isEnabled() )
      return;

   m_Enabled = true;
   buildModel();
}


void RveCartesianModel::disable()
{
   if( !isEnabled() )
      return;

   m_Enabled = false;
   clearModel();
}


bool RveCartesianModel::isEnabled() const
{
   return( m_Enabled );
}


void RveCartesianModel::clearModel()
{
   clearVertexModel();
   clearGLModel();
}


void RveCartesianModel::paintGL()
{
   if( !isEnabled() )
      return;

   GLuint indices = m_GLIndices;
   GLuint vertices = m_GLVertices;
   int numVertices = m_Indices.size();

   glBindBuffer( GL_ARRAY_BUFFER, vertices );
   glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, indices );
   glEnableClientState( GL_VERTEX_ARRAY );
   glEnableClientState( GL_COLOR_ARRAY );
   glColorPointer( 4, GL_FLOAT, sizeof( Vertex ), (char *)NULL + 12 );
   glVertexPointer( 3, GL_FLOAT, sizeof( Vertex ), (char *)NULL + 0 );

   glDrawElements( GL_TRIANGLES, numVertices, GL_UNSIGNED_INT, (char *)NULL + 0 );

   glDisableClientState( GL_VERTEX_ARRAY );
   glDisableClientState( GL_COLOR_ARRAY );
}


void RveCartesianModel::clearGLModel()
{
   glDeleteBuffers( 1, &m_GLIndices );
   glDeleteBuffers( 1, &m_GLVertices );

   m_GLIndices = 0;
   m_GLVertices = 0;
}


void RveCartesianModel::buildGLModel()
{
   clearGLModel();

   GLuint glindices;
   glGenBuffers( 1, &glindices );
   glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, glindices );
   glBufferData( GL_ELEMENT_ARRAY_BUFFER,
                  m_Indices.size() * sizeof( int ),
                  m_Indices.data(),
                  GL_STATIC_DRAW );
   glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
   m_GLIndices = glindices ;

   GLuint glcolors;
   glGenBuffers( 1, &glcolors );
   glBindBuffer( GL_ARRAY_BUFFER, glcolors );
   glBufferData( GL_ARRAY_BUFFER,
                  m_Vertices.size() * sizeof( Vertex ),
                  m_Vertices.data(),
                  GL_STATIC_DRAW );
   glBindBuffer( GL_ARRAY_BUFFER, 0 );
   m_GLVertices = glcolors;
}


double RveCartesianModel::getValue( double dh, double dx, double dy )
{
   double range = sqrt( (dx * dx) + (dy * dy) + (dh * dh) );
   double realAzimuth = (atan2( dy, dx ) * 360.0) / (2.0 * M_PI);
   realAzimuth = 90.0 - realAzimuth;

   while( realAzimuth < 0.0 )
      realAzimuth += 360.0;
   realAzimuth = fmod( realAzimuth, 360.0 );

   double realElevation = 90.0;
   if( range != 0.0 )
   {
      realElevation = (asin( dh / range ) * 360.0) / (2.0 * M_PI);
   }
   double elUpper = -DBL_MAX;
   double elLower = -DBL_MAX;
   double scaleUpper = -DBL_MAX;
   m_pRawdata->nearestElevation( realElevation, dh, elLower, elUpper, scaleUpper );
   if( scaleUpper < 0.0 )
      return( -DBL_MAX );
   double scaleLower = 1.0 - scaleUpper;
   double azUpper = m_pRawdata->nearestAzimuth( elUpper, realAzimuth );
   double azLower = m_pRawdata->nearestAzimuth( elLower, realAzimuth );

   QMap<double, QVector<double> > *sliceUpper = m_pRawdata->slice( elUpper );
   QMap<double, QVector<double> > *sliceLower = m_pRawdata->slice( elLower );
   QVector<double> &rayUpper = (*sliceUpper)[azUpper];
   QVector<double> &rayLower = (*sliceLower)[azLower];
   double rstepUpper = m_pRawdata->rangeStep( elUpper );
   double rstepLower = m_pRawdata->rangeStep( elLower );
   double startRangeUpper = m_pRawdata->startRange( elUpper );
   double startRangeLower = m_pRawdata->startRange( elLower );

   double rangeUpper = dh / sin( (elUpper / 360.0) * 2.0 * M_PI );
   double rangeLower = dh / sin( (elLower / 360.0) * 2.0 * M_PI );

   int indexUpper = ((range - startRangeUpper) / rstepUpper) + 0.5;
   int indexLower = ((range - startRangeLower) / rstepLower) + 0.5;
//   int indexUpper = ((rangeUpper - startRangeUpper) / rstepUpper) + 0.5;
//   int indexLower = ((rangeLower - startRangeLower) / rstepLower) + 0.5;

   bool vUpperValid = false;
   bool vLowerValid = false;
   double vUpper;
   double vLower;

   if( indexUpper >= 0 && indexUpper < rayUpper.size() )
   {
      vUpper = rayUpper[indexUpper];
      if( vUpper != -DBL_MAX )
         vUpperValid = true;
   }

   if( indexLower >= 0 && indexLower < rayLower.size() )
   {
      vLower = rayLower[indexLower];
      if( vLower != -DBL_MAX )
         vLowerValid = true;
   }

   if( vLowerValid && vUpperValid )
   {
//      return( scaleUpper >= 0.5 ? vUpper : vLower );
      return( (vUpper * scaleUpper) + (vLower * scaleLower) );
   } else
   {
      return( -DBL_MAX );
   }
/*
   int rIndex = ((range - startRange) / rstep) + 0.5;

   if( rIndex < 0 || rIndex >= ray.size() )
   {
      return( -DBL_MAX );
   } else
   {
      double v = ray[rIndex];
      return( v );
   }*/
   return( -DBL_MAX );
}


void RveCartesianModel::buildVertexModel( double dh, bool keepData )
{
   double range = m_pRawdata->range();

   QVector<Vertex> &vertices = m_Vertices;
   QVector<double> &data = m_Data;
   QVector<int> &indices = m_Indices;

   int startOffset = m_Vertices.size();

   for( int y = 0; y < parameters().cellsPerLayer(); y++ )
   {
      double dy = ( ( 2.0 * range ) * ( (double)y / (double)( parameters().cellsPerLayer() - 1 ) ) ) - range;

      for( int x = 0; x < parameters().cellsPerLayer(); x++ )
      {
         double dx = ( m_Width * ( (double)x / (double)( parameters().cellsPerLayer() - 1 ) ) ) - ( m_Width / 2.0 );
         double value = getValue( dh, dx, dy );
         double min = parameters().clipMin();
         double max = parameters().clipMax();
         bool nodata = false;

         if( keepData )
         {
            data.append( value );
         }

         if( value != -DBL_MAX )
         {
            if( value < min )
            {
               nodata = true;
            } else
            if ( value > max )
            {
               value = max;
            }
         } else
         {
            nodata = true;
         }

         if( value != -DBL_MAX && !nodata )
         {
               value = (value - min) / (max - min);
               //value = 1.0;
         } else
         {
            value = -1.0;
         }

         Vertex v;
         v.color.r = value;
         v.color.g = value;
         v.color.b = value;
         v.color.a = 1.0;
         v.point.x = dx;
         v.point.z = dh;
         v.point.y = dy;
         vertices.append( v );

//         img.setPixel( x, y, qRgb( value * 255, value * 255, value * 255 ) );
      }
   }

   for( int y = 0; y < parameters().cellsPerLayer() - 1; y++ )
   {
      for( int x = 0; x < parameters().cellsPerLayer() - 1; x++ )
      {
         double value = vertices[startOffset + (parameters().cellsPerLayer() * y) + x].color.r;

         if( value > 0.0 )
         {
            indices.append( startOffset + ((y + 0) * parameters().cellsPerLayer()) + (x + 0) );
            indices.append( startOffset + ((y + 0) * parameters().cellsPerLayer()) + (x + 1) );
            indices.append( startOffset + ((y + 1) * parameters().cellsPerLayer()) + (x + 0) );

            indices.append( startOffset + ((y + 0) * parameters().cellsPerLayer()) + (x + 1) );
            indices.append( startOffset + ((y + 1) * parameters().cellsPerLayer()) + (x + 1) );
            indices.append( startOffset + ((y + 1) * parameters().cellsPerLayer()) + (x + 0) );
         }
      }
   }
}


QVector<double> &RveCartesianModel::data()
{
   return( m_Data );
}


void RveCartesianModel::buildVertexModel()
{
   buildVertexModel( false );
}


void RveCartesianModel::buildVertexModel( bool keepData )
{
   double lowest = parameters().lowestLayer();
   double highest = parameters().highestLayer();

   for( int layer = 0; layer < parameters().layers(); layer++ )
   {
      double dh = lowest + ( ( highest - lowest ) * ( (double)layer / (double)( parameters().layers() - 1 ) ) );
      buildVertexModel( dh, keepData );
   }
}


void RveCartesianModel::clearVertexModel()
{
   m_Vertices.clear();
   m_Data.clear();
   m_Indices.clear();
}


void RveCartesianModel::buildModel()
{
   buildModel( m_KeepData );
}


void RveCartesianModel::buildModel( bool keepData )
{
   clearModel();

   if( !isEnabled() )
      return;
   if( !m_pRawdata )
      return;
   if( !m_pRawdata->isLoaded() )
      return;
   if( !m_Parameters.valid() )
      return;

   buildVertexModel( keepData );
   buildGLModel();
}


double RveCartesianModel::layerWidth() const
{
   return( m_Width );
}


QVector<RveModel::Vertex> &RveCartesianModel::vertices()
{
   return( m_Vertices );
}


void RveCartesianModel::setRawdata( RveRawdata *pRawdata )
{
   if( m_pRawdata != pRawdata )
   {
      m_pRawdata = pRawdata;
      m_Width = m_pRawdata->range() * 2.0;
      if( isEnabled() )
         buildModel();
   }
}


RveCartesianParameters RveCartesianModel::parameters() const
{
   return( m_Parameters );
}


void RveCartesianModel::setParameters( RveCartesianParameters parameters )
{
   if( m_Parameters != parameters )
   {
      m_Parameters = parameters;
      if( isEnabled() )
         buildModel();
   }
}
