/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/

#define _USE_MATH_DEFINES
#include <math.h>

#include "RveFlatModel.h"

#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE  0x809D
#endif

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveFlatModel::RveFlatModel() :
   RveModel( 0 ),
   m_Enabled( false ),
   m_Parameters( RveFlatParameters() )
{
   initializeGLFunctions();
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveFlatModel::~RveFlatModel()
{
   clearModel();
}


void RveFlatModel::enable()
{
   if( isEnabled() )
      return;

   m_Enabled = true;
   buildModel();
}


void RveFlatModel::disable()
{
   if( !isEnabled() )
      return;

   m_Enabled = false;
   clearModel();
}


bool RveFlatModel::isEnabled() const
{
   return( m_Enabled );
}


void RveFlatModel::setRawdata( RveRawdata *pRawdata )
{
   if( m_pRawdata != pRawdata )
   {
      m_pRawdata = pRawdata;
      if( isEnabled() )
         buildModel();
   }
}


void RveFlatModel::setParameters( RveFlatParameters parameters )
{
   if( m_Parameters != parameters )
   {
      m_Parameters = parameters;
      if( isEnabled() )
         buildModel();
   }
}


RveFlatParameters RveFlatModel::parameters() const
{
   return( m_Parameters );
}


void RveFlatModel::paintGL()
{
   if( !isEnabled() )
      return;

   foreach( double elevation, m_Indices.keys() )
   {
      GLuint indices = m_GLIndices[elevation];
      GLuint vertices = m_GLVertices[elevation];
      int numVertices = m_Indices[elevation].size();

      glBindBuffer( GL_ARRAY_BUFFER, vertices );
      glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, indices );
      glEnableClientState( GL_VERTEX_ARRAY );
      glEnableClientState( GL_COLOR_ARRAY );
      glColorPointer( 4, GL_FLOAT, sizeof( Vertex ), (char *)NULL + 12 );
      glVertexPointer( 3, GL_FLOAT, sizeof( Vertex ), (char *)NULL + 0 );

      glDrawElements( GL_TRIANGLES, numVertices, GL_UNSIGNED_INT, (char *)NULL + 0 );

      glDisableClientState( GL_VERTEX_ARRAY );
      glDisableClientState( GL_COLOR_ARRAY );
   }
}


void RveFlatModel::clearGLModel()
{
   foreach( double elevation, m_GLIndices.keys() )
   {
      glDeleteBuffers( 1, &m_GLIndices[elevation] );
   }

   foreach( double elevation, m_GLVertices.keys() )
   {
      glDeleteBuffers( 1, &m_GLVertices[elevation] );
   }

   m_GLIndices.clear();
   m_GLVertices.clear();
}


void RveFlatModel::buildGLModel()
{
   clearGLModel();

   foreach( double elevation, m_Indices.keys() )
   {
      GLuint glindices;
      glGenBuffers( 1, &glindices );
      glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, glindices );
      glBufferData( GL_ELEMENT_ARRAY_BUFFER,
                    m_Indices[elevation].size() * sizeof( int ),
                    m_Indices[elevation].data(),
                    GL_STATIC_DRAW );
      glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
      m_GLIndices.insert( elevation, glindices );

      GLuint glcolors;
      glGenBuffers( 1, &glcolors );
      glBindBuffer( GL_ARRAY_BUFFER, glcolors );
      glBufferData( GL_ARRAY_BUFFER,
                    m_Vertices[elevation].size() * sizeof( Vertex ),
                    m_Vertices[elevation].data(),
                    GL_STATIC_DRAW );
      glBindBuffer( GL_ARRAY_BUFFER, 0 );
      m_GLVertices.insert( elevation, glcolors );
   }
}


void RveFlatModel::buildVertexModel( double elevation )
{
   double rangeStep = m_pRawdata->rangeStep( elevation);
   double rangeStart = m_pRawdata->startRange( elevation );
   double rangeStop = m_pRawdata->stopRange( elevation );
   int    numBins = m_pRawdata->numBins( elevation );

   QMap<double, QVector<double> > *slice = m_pRawdata->slice( elevation );
   QVector<double> azimuthAngles = slice->keys().toVector();
   int numRays = azimuthAngles.size();

   m_Vertices[elevation].clear();
   m_Indices[elevation].clear();

   QVector<Vertex> &vertices = m_Vertices[elevation];
   QVector<int> &indices = m_Indices[elevation];

   // Build colors and coordinates
   for( int r = 0; r < numRays; r++ )
   {
      double azimuth = azimuthAngles[r];
      double rAzimuth = 90.0 - azimuth;
      double azimuthRad = rAzimuth * 2.0 * M_PI / 360.0;
      const QVector<double> &ray = (*slice)[azimuth];

      for( int b = 0; b < numBins; b++ )
      {
         double value = ray[b];

         double range = rangeStart + ( rangeStep * (double)b );
         double height = m_pRawdata->beamHeight( range, elevation );
         Vertex v;
         v.point.z = height;
         v.point.x = range * cos( azimuthRad );
         v.point.y = range * sin( azimuthRad );

         bool nodata = false;

         if( value != -DBL_MAX )
         {
            if( value < m_Parameters.clipMin() )
            {
               nodata = true;
            } else
            if( value > m_Parameters.clipMax() )
            {
               value = m_Parameters.clipMax();
            }
         }

         if( value != -DBL_MAX && !nodata)
         {
            value = (value - m_Parameters.clipMin()) / (m_Parameters.clipMax() - m_Parameters.clipMin());
         } else
         {
            value = -1;
         }

         v.color.r = value;
         v.color.g = value;
         v.color.b = value;
         v.color.a = 1.0;

         vertices.append( v );
      }
   }

   // Build indices
   for( int r = 0; r < numRays; r++ )
   {
//      double azimuth = azimuthAngles[r];
//      const QVector<double> &ray = (*slice)[azimuth];

      for( int b = 0; b < numBins - 1; b++ )
      {
//         double value = ray[b];
         double value = vertices[r * numBins + b].color.r;

         if( value >= 0 )
         {
            // First triangle
            indices.append( (((r + 0) % numRays) * numBins) + (b + 0) );
            indices.append( (((r + 0) % numRays) * numBins) + (b + 1) );
            indices.append( (((r + 1) % numRays) * numBins) + (b + 0) );
            // Second triangle
            indices.append( (((r + 1) % numRays) * numBins) + (b + 0) );
            indices.append( (((r + 0) % numRays) * numBins) + (b + 1) );
            indices.append( (((r + 1) % numRays) * numBins) + (b + 1) );
         }
      }
   }
}


void RveFlatModel::clearModel()
{
   clearVertexModel();
   clearGLModel();
}


void RveFlatModel::clearVertexModel()
{
   m_Vertices.clear();
   m_Indices.clear();
}


void RveFlatModel::buildVertexModel()
{
   if( !m_pRawdata->isLoaded() )
      return;

   clearVertexModel();

   foreach( double elevation, m_pRawdata->elevations() )
   {
      buildVertexModel( elevation );
   }
}


void RveFlatModel::buildModel()
{
   clearModel();

   if( !isEnabled() )
      return;
   if( !m_pRawdata )
      return;
   if( !m_pRawdata->isLoaded() )
      return;
   if( !m_Parameters.valid() )
      return;

   buildVertexModel();
   buildGLModel();
}
