/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/

#include <QFile>
#include <QSet>

#include "RveDbfFile.h"

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveDbfFile::RveDbfFile( const QString &fileName, QObject *parent ) :
   QObject( parent ),
   m_FileName( fileName ),
   m_Valid( false )
{
   m_Valid = load( fileName );
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveDbfFile::~RveDbfFile()
{
   qDeleteAll( m_Records );
   qDeleteAll( m_Fields );

   m_Records.clear();
   m_Fields.clear();
}


bool RveDbfFile::load( const QString &fileName )
{
   QFile f( fileName );
   if( !f.open( QIODevice::ReadOnly ) )
   {
      return( false );
   }

   qint64 fileSize = f.size();
   QByteArray header = f.read( 32 );
   if( header.size() != 32 )
   {
      f.close();
      return( false );
   }

   m_Version = header[0];
   if( m_Version != 3 )
   {
      f.close();
      return( false );
   }

   QDate m_LastUpdate = QDate( header[1] + 1900, header[2], header[3] );
   quint32 numRecords = toUint32( header, 4, false );

   quint16 headerLength = toUint16( header, 8, false );
   m_RecordLength = toUint16( header, 10, false );

   int numFields = ( headerLength - 32 - 1 ) / 32;

   for( int i = 0; i < numFields; i++ )
   {
      QByteArray field = f.read( 32 );
      RveDbfField *pField = new RveDbfField( field );
      if( pField->isValid() )
      {
         m_Fields.append( pField );
      } else
      {
         f.close();
         delete pField;
         return( false );
      }
   }

   quint8 tmp = 0x00;
   f.read( (char*)&tmp, 1 );
   if( tmp != 0x0d )
   {
      f.close();
      return( false );
   }

   f.seek( headerLength );

   for( int i = 0; i < numRecords; i++ )
   {
      QByteArray record = f.read( m_RecordLength );
      if( record.size() != m_RecordLength )
      {
         f.close();
         return( false );
      }

      RveDbfRecord *pDbfRecord = new RveDbfRecord( record, this );
      if( pDbfRecord->isValid() )
      {
         m_Records.append( pDbfRecord );
      } else
      {
         delete pDbfRecord;
      }
   }

   f.close();

   return( true );
}


double RveDbfFile::toDouble( const QByteArray &p, int offset, bool bigEndian )
{
   double result = 0.0;
   quint64 *dv = (quint64 *)&result;

   if( bigEndian )
   {
      *dv = (quint64)
         ( (quint64)p[offset + 7] << 0  ) & 0x00000000000000ff |
         ( (quint64)p[offset + 6] << 8  ) & 0x000000000000ff00 |
         ( (quint64)p[offset + 5] << 16 ) & 0x0000000000ff0000 |
         ( (quint64)p[offset + 4] << 24 ) & 0x00000000ff000000 |
         ( (quint64)p[offset + 3] << 32 ) & 0x000000ff00000000 |
         ( (quint64)p[offset + 2] << 40 ) & 0x0000ff0000000000 |
         ( (quint64)p[offset + 1] << 48 ) & 0x00ff000000000000 |
         ( (quint64)p[offset + 0] << 56 ) & 0xff00000000000000;
   } else
   {
      *dv = (quint64)
         ( (quint64)p[offset + 0] << 0  ) & 0x00000000000000ff |
         ( (quint64)p[offset + 1] << 8  ) & 0x000000000000ff00 |
         ( (quint64)p[offset + 2] << 16 ) & 0x0000000000ff0000 |
         ( (quint64)p[offset + 3] << 24 ) & 0x00000000ff000000 |
         ( (quint64)p[offset + 4] << 32 ) & 0x000000ff00000000 |
         ( (quint64)p[offset + 5] << 40 ) & 0x0000ff0000000000 |
         ( (quint64)p[offset + 6] << 48 ) & 0x00ff000000000000 |
         ( (quint64)p[offset + 7] << 56 ) & 0xff00000000000000;
   }

   return( result );
}


quint32 RveDbfFile::toUint32( const QByteArray &p, int offset, bool bigEndian )
{
   quint32 result = 0;

   if( bigEndian )
   {
      result =
         ( ( p[offset + 3] << 0  ) & 0x000000ff |
           ( p[offset + 2] << 8  ) & 0x0000ff00 |
           ( p[offset + 1] << 16 ) & 0x00ff0000 |
           ( p[offset + 0] << 24 ) & 0xff000000 );
   } else
   {
      result =
         ( ( p[offset + 0] << 0  ) & 0x000000ff |
           ( p[offset + 1] << 8  ) & 0x0000ff00 |
           ( p[offset + 2] << 16 ) & 0x00ff0000 |
           ( p[offset + 3] << 24 ) & 0xff000000 );
   }

   return( result );
}


quint16 RveDbfFile::toUint16( const QByteArray &p, int offset, bool bigEndian )
{
   quint16 result = 0;

   if( bigEndian )
   {
      result =
         ( ( p[offset + 1] << 0 ) & 0x00ff ) |
         ( ( p[offset + 0] << 8 ) & 0xff00 );
   } else
   {
      result =
         ( ( p[offset + 0] << 0 ) & 0x00ff ) |
         ( ( p[offset + 1] << 8 ) & 0xff00 );
   }

   return( result );
}

QString RveDbfFile::fileName() const
{
   return( m_FileName );
}


bool RveDbfFile::isValid() const
{
   return( m_Valid );
}


int RveDbfFile::numFields() const
{
   return( m_Fields.size() );
}


RveDbfField *RveDbfFile::field( int n ) const
{
   if( n < 0 || n >= m_Fields.size() )
      return( 0 );

   return( m_Fields[n] );
}


RveDbfField *RveDbfFile::field( const QString &fieldName, bool caseSensitive ) const
{
   return( field( indexOfField( fieldName, caseSensitive ) ) );
}


int RveDbfFile::numRecords() const
{
   return( m_Records.size() );
}


const RveDbfRecord *RveDbfFile::record( int n ) const
{
   if( n < 0 || n >= m_Records.size() )
      return( 0 );

   return( m_Records[n] );
}


quint16 RveDbfFile::recordLength() const
{
   return( m_RecordLength );
}


int RveDbfFile::indexOfField( const QString &name, bool caseSensitive ) const
{
   for( int i = 0; i < m_Fields.size(); i++ )
   {
      if( caseSensitive )
      {
         if( m_Fields[i]->name().trimmed() == name.trimmed() )
            return( i );
      } else
      {
         if( m_Fields[i]->name().trimmed().toLower() == name.trimmed().toLower() )
            return( i );
      }
   }

   return( -1 );
}


int RveDbfFile::numDistinctiveRecords( int fieldNum ) const
{
   QSet<QString> values;

   for( int i = 0; i < numRecords(); i++ )
   {
      QVariant v = record( i )->value( fieldNum );
      QString sv = v.toString();
      if( !values.contains( sv ) )
         values.insert( sv );
   }

   return( values.size() );
}