/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/
#include <QDir>
#include <QFileInfo>

#include "RveGisData.h"

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveGisData::RveGisData( const QString &fileName, QObject *parent ) :
   QObject( parent ),
   m_Valid( false ),
   m_pShpFile( 0 ),
   m_pDbfFile( 0 ),
   m_Enabled( true ),
   m_Font( QFont( "Arial", 10 ) ),
   m_FontScale( 250.0 )
{
   m_Valid = load( fileName );
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveGisData::~RveGisData()
{
   if( m_pShpFile )
   {
      delete m_pShpFile;
      m_pShpFile = 0;
   }
   
   if( m_pDbfFile )
   {
      delete m_pDbfFile;
      m_pDbfFile = 0;
   }
}


bool RveGisData::load( const QString &fileName )
{
   QString shpFileName;
   QString dbfFileName;

   m_Label = "";
   QString fn = fileName;

   if( QFileInfo( fn ).suffix().toLower() == "shp" )
      shpFileName = fn;
   else
   if( QFileInfo( fn ).suffix().toLower() == "dbf" )
      dbfFileName = fn;

   if( shpFileName.isEmpty() || dbfFileName.isEmpty() )
   {
      QFileInfo fInfo( fn );
      QString path = fInfo.absolutePath();
      QDir dir( path );
      QStringList lst;
      if( shpFileName.isEmpty() )
         lst << "*.shp";
      if( dbfFileName.isEmpty() )
         lst << "*.dbf";
      QFileInfoList fileInfoList = dir.entryInfoList( lst, QDir::Files | QDir::Readable );
      foreach( QFileInfo f, fileInfoList )
      {
         if( f.suffix().toLower() == "dbf"  &&
             f.baseName() == QFileInfo( fn ).baseName() )
            dbfFileName = f.absoluteFilePath();
         else
         if( f.suffix().toLower() == "shp" &&
             f.baseName() == QFileInfo( fn ).baseName() )
            shpFileName = f.absoluteFilePath();
      }
   }

   m_FileName = fn;

   m_pShpFile = new RveShpFile( shpFileName, this );
   if( !m_pShpFile->isValid() )
   {
      delete m_pShpFile;
      m_pShpFile = 0;
   } else
   {
      connect( m_pShpFile, SIGNAL( changed() ), this, SIGNAL( changed() ) );
   }

   m_pDbfFile = new RveDbfFile( dbfFileName, this );
   if( !m_pDbfFile->isValid() )
   {
      delete m_pDbfFile;
      m_pDbfFile = 0;
   } else
   {
      if( m_pDbfFile->numFields() > 0 )
      {
         m_Label = m_pDbfFile->field( 0 )->name();
      }
   }
   return( m_pShpFile != 0 );
}


QString RveGisData::fileName() const
{
   return( m_FileName );
}


QColor RveGisData::color() const
{
   return( m_pShpFile->color() );
}


void RveGisData::setColor( const QColor &color )
{
   m_pShpFile->setColor( color );
}


bool RveGisData::isValid() const
{
   return( m_Valid );
}


int RveGisData::lineWidth() const
{
   return( m_pShpFile->lineWidth() );
}


void RveGisData::setLineWidth( int lineWidth )
{
   m_pShpFile->setLineWidth( lineWidth );
}


bool RveGisData::isEnabled() const
{
   return( m_Enabled );
}


void RveGisData::setEnabled( bool e )
{
   m_Enabled = e;

   emit changed();
}


void RveGisData::calculate( const RveProjPnt &projPnt, double range_m, QVector<RveModel::Vertex> &map, int precision )
{
   if( isEnabled() )
   {
      m_pShpFile->calculate( projPnt, range_m, map, precision );
   }
}


void RveGisData::setFontScale( double scale )
{
   bool doUpdate = m_FontScale != scale;
   m_FontScale = scale;

   if( doUpdate )
      emit changed();
}


double RveGisData::fontScale() const
{
   return( m_FontScale );
}


QFont RveGisData::font() const
{
   return( m_Font );
}


void RveGisData::setFont( const QFont &font )
{
   bool doUpdate = m_Font != font;
   m_Font = font;

   if( doUpdate )
      emit changed();
}


void RveGisData::clearGL()
{
   for( int i = 0; i < m_GLVertices.size(); i++ )
   {
      glDeleteBuffers( 1, &m_GLVertices[i] );
   }
   m_GLVertices.clear();
   m_GLNumVertices.clear();
}


const RveShpFile *RveGisData::shpFile() const
{
   return( m_pShpFile );
}


const RveDbfFile *RveGisData::dbfFile() const
{
   return( m_pDbfFile );
}


QString RveGisData::label() const
{
   return( m_Label );
}


void RveGisData::setLabel( const QString &label )
{
   m_Label = label;

   emit changed();
}


QString RveGisData::filterFieldName() const
{
   return( m_FilterFieldName );
}


void RveGisData::setFilterFieldName( const QString &name )
{
   m_FilterFieldName = name;
}


bool RveGisData::valueFiltered( const QVariant &value ) const
{
   return( m_FilterValues.contains( value ) );
}


void RveGisData::setValueFiltered( const QVariant &value, bool add )
{
   m_FilterValues.removeAll( value );
   if( add )
      m_FilterValues.append( value );
}


void RveGisData::clearFilterValues()
{
   m_FilterValues.clear();
}


bool RveGisData::valueFiltered( int n ) const
{
   if( m_FilterFieldName.isEmpty() )
      return( true );
   int fieldIndex = m_pDbfFile->indexOfField( m_FilterFieldName );
   if( fieldIndex < 0 )
      return( true );

   return( m_FilterValues.contains( m_pDbfFile->record( n )->value( fieldIndex ) ) );
}
