/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/

#include <QFile>
#include <QImage>
#include <QPainter>

#include "RveShpFile.h"
#include "RveShp.h"
#include "RveShpPolygon.h"
#include "RveShpPoint.h"

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveShpFile::RveShpFile( const QString &fileName, QObject *parent ) :
   QObject( parent ),
   m_Color( QColor( 255, 255, 255, 255 ) ),
   m_LineWidth( 1 ),
   m_Enabled( true ),
   m_Valid( false )
{
   m_Valid = load( fileName );
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveShpFile::~RveShpFile()
{
   qDeleteAll( m_Shapes );
   m_Shapes.clear();
}


bool RveShpFile::load( const QString &fileName )
{
   QFile f( fileName );
   if( !f.open( QIODevice::ReadOnly ) )
   {
      return( false );
   }

   qint64 fileSize = f.size();
   QByteArray header = f.read( 100 );
   if( header.size() != 100 )
   {
      f.close();
      return( false );
   }

   m_FileCode = toUint32( header, 0, true );
   if( m_FileCode != 9994 )
   {
      f.close();
      return( false );
   }

   m_Version = toUint32( header, 28, false );
   if( m_Version != 1000 )
   {
      f.close();
      return( false );
   }

   quint32 shpFileLength = toUint32( header, 24, true ) * 2;
   if( shpFileLength != fileSize )
   {
      f.close();
      return( false );
   }

   m_ShapeType = (SHP_TYPE)toUint32( header, 32, false );

   m_xMin = toDouble( header, 36, false );
   m_yMin = toDouble( header, 44, false );
   m_xMax = toDouble( header, 52, false );
   m_yMax = toDouble( header, 60, false );

   while( !f.atEnd() )
   {
      header = f.read( 8 );
      if( header.size() != 8 )
      {
         break;
      }

      quint32 recordNumber = toUint32( header, 0, true );
      quint32 contentLength = toUint32( header, 4, true ) * 2;

      QByteArray content = f.read( contentLength );
      if( content.size() != contentLength )
      {
         break;
      }

      RveShp::SHP_TYPE shpType = (RveShp::SHP_TYPE)toUint32( content, 0, false );
      if( shpType == RveShp::TYPE_POLYGON )
      {
         RveShpPolygon *pShpPolygon = new RveShpPolygon( content, recordNumber, this );
         if( pShpPolygon->valid() )
         {
            m_Shapes.append( pShpPolygon );
         } else
         {
            delete pShpPolygon;
         }
      } else
      if( shpType == RveShp::TYPE_POINT )
      {
         RveShpPoint *pShpPoint = new RveShpPoint( content, recordNumber, this );
         if( pShpPoint->valid() )
         {
            m_Shapes.append( pShpPoint );
         } else
         {
            delete pShpPoint;
         }
      } else
      {
         printf( "\n" );
      }
   }
   f.close();

   return( true );
}


double RveShpFile::toDouble( const QByteArray &p, int offset, bool bigEndian )
{
   double result = 0.0;
   quint64 *dv = (quint64 *)&result;

   if( bigEndian )
   {
      *dv = (quint64)
         ( (quint64)p[offset + 7] << 0  ) & 0x00000000000000ff |
         ( (quint64)p[offset + 6] << 8  ) & 0x000000000000ff00 |
         ( (quint64)p[offset + 5] << 16 ) & 0x0000000000ff0000 |
         ( (quint64)p[offset + 4] << 24 ) & 0x00000000ff000000 |
         ( (quint64)p[offset + 3] << 32 ) & 0x000000ff00000000 |
         ( (quint64)p[offset + 2] << 40 ) & 0x0000ff0000000000 |
         ( (quint64)p[offset + 1] << 48 ) & 0x00ff000000000000 |
         ( (quint64)p[offset + 0] << 56 ) & 0xff00000000000000;
   } else
   {
      *dv = (quint64)
         ( (quint64)p[offset + 0] << 0  ) & 0x00000000000000ff |
         ( (quint64)p[offset + 1] << 8  ) & 0x000000000000ff00 |
         ( (quint64)p[offset + 2] << 16 ) & 0x0000000000ff0000 |
         ( (quint64)p[offset + 3] << 24 ) & 0x00000000ff000000 |
         ( (quint64)p[offset + 4] << 32 ) & 0x000000ff00000000 |
         ( (quint64)p[offset + 5] << 40 ) & 0x0000ff0000000000 |
         ( (quint64)p[offset + 6] << 48 ) & 0x00ff000000000000 |
         ( (quint64)p[offset + 7] << 56 ) & 0xff00000000000000;
   }

   return( result );
}


quint32 RveShpFile::toUint32( const QByteArray &p, int offset, bool bigEndian )
{
   quint32 result = 0;

   if( bigEndian )
   {
      result =
         ( ( p[offset + 3] << 0  ) & 0x000000ff |
           ( p[offset + 2] << 8  ) & 0x0000ff00 |
           ( p[offset + 1] << 16 ) & 0x00ff0000 |
           ( p[offset + 0] << 24 ) & 0xff000000 );
   } else
   {
      result =
         ( ( p[offset + 0] << 0  ) & 0x000000ff |
           ( p[offset + 1] << 8  ) & 0x0000ff00 |
           ( p[offset + 2] << 16 ) & 0x00ff0000 |
           ( p[offset + 3] << 24 ) & 0xff000000 );
   }

   return( result );
}


void RveShpFile::calculate( const RveProjPnt &projPnt, double range_m, QVector<RveModel::Vertex> &map, int precision )
{
   if( !isEnabled() )
      return;

   QImage img( precision, precision, QImage::Format_ARGB32 );
   img.fill( QColor( 0, 0, 0 ) );
   QPainter painter( &img );
   QPen pen( QColor( 255, 255, 255 ) );
   pen.setWidth( m_LineWidth );
   painter.setPen( pen );

   foreach( RveShp *pShp, m_Shapes )
   {
      pShp->calculate( painter, projPnt, range_m, precision );
   }

   for( int y = 0; y < img.width(); y++ )
   {
      for( int x = 0; x < img.height(); x++ )
      {
         QRgb v = img.pixel( x, y );
         if( v & 0xffffff )
         {
            map[( y * precision ) + x].color.r = (double)m_Color.red() / 255.0;
            map[( y * precision ) + x].color.g = (double)m_Color.green() / 255.0;
            map[( y * precision ) + x].color.b = (double)m_Color.blue() / 255.0;
            map[( y * precision ) + x].color.a = 1.0;
         }
      }
   }
}


QString RveShpFile::fileName() const
{
   return( m_FileName );
}


QColor RveShpFile::color() const
{
   return( m_Color );
}


int RveShpFile::lineWidth() const
{
   return( m_LineWidth );
}


void RveShpFile::setLineWidth( int lineWidth )
{
   m_LineWidth = lineWidth;

   emit changed();
}


bool RveShpFile::isEnabled() const
{
   return( m_Enabled );
}


void RveShpFile::setEnabled( bool e )
{
   m_Enabled = e;

   emit changed();
}


void RveShpFile::setColor( const QColor &color )
{
   m_Color = color;

   emit changed();
}


bool RveShpFile::isValid() const
{
   return( m_Valid );
}


RveShpFile::SHP_TYPE RveShpFile::type() const
{
   return( m_ShapeType );
}


int RveShpFile::numShapes() const
{
   return( m_Shapes.size() );
}


const RveShp *RveShpFile::shape( int n ) const
{
   if( n < 0 || n >= m_Shapes.size() )
      return( 0 );

   return( m_Shapes[n] );
}
