/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/

#include <QFile>
#include <QImage>
#include <QPainter>

#include "RveShpFile.h"
#include "RveShpPolygon.h"

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveShpPolygon::RveShpPolygon( const QByteArray &p, quint32 recordNumber, QObject *parent ) :
   RveShp( recordNumber, parent )
{
   m_Valid = load( p );
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveShpPolygon::~RveShpPolygon()
{
}


bool RveShpPolygon::load( const QByteArray &p )
{
   int contentLength = p.size();

   m_ShapeType = (RveShp::SHP_TYPE)RveShpFile::toUint32( p, 0, false );
   if( m_ShapeType != RveShp::TYPE_POLYGON )
      return( false );

   m_Xmin = RveShpFile::toDouble( p, 4, false );
   m_Ymin = RveShpFile::toDouble( p, 12, false );
   m_Xmax = RveShpFile::toDouble( p, 20, false );
   m_Ymax = RveShpFile::toDouble( p, 28, false );

   quint32 numParts = RveShpFile::toUint32( p, 36, false );
   quint32 numTotalPoints = RveShpFile::toUint32( p, 40, false );

   QVector<quint32> parts;
   for( quint32 i = 0; i < numParts; i++ )
   {
      quint32 pt = RveShpFile::toUint32( p, 44 + ( i * 4 ), false );
      parts.append( pt );
   }

   QVector<quint32> numPoints;
   for( quint32 curPart = 0; curPart < numParts; curPart++ )
   {
      quint32 curNumPoints;
      if( curPart == numParts - 1 )
      {
         curNumPoints = numTotalPoints - parts[curPart];
      } else
      {
         curNumPoints = parts[curPart + 1] - parts[curPart];
      }

      numPoints.append( curNumPoints );
   }

   for( quint32 curPart = 0; curPart < numParts; curPart++ )
   {
      QVector<Point> points;
      for( quint32 curPoint = 0; curPoint < numPoints[curPart]; curPoint++ )
      {
         quint32 curTotalPoint = parts[curPart] + curPoint;
         quint32 startOffset = 44 + ( 4 * numParts ) + ( 2 * 8 * curTotalPoint );
         Point point;

         point.lon = RveShpFile::toDouble( p, startOffset + 0, false );
         point.lat = RveShpFile::toDouble( p, startOffset + 8, false );

         points.append( point );
      }
      m_Parts.append( points );
   }

   return( true );
}


int RveShpPolygon::numParts() const
{
   return( m_Parts.size() );
}

const QVector<RveShp::Point> *RveShpPolygon::part( int n ) const
{
   if( n < 0 || n >= m_Parts.size() )
      return( 0 );

   return( &m_Parts[n] );
}


void RveShpPolygon::calculate( QPainter &painter, const RveProjPnt &projPnt, double range_m, int precision )
{
   double width = range_m * 2.0;
   double metersPerCell = width / (double)precision;

   for( int i = 0; i < numParts(); i++ )
   {
      const QVector<RveShp::Point> *pPolygon = part( i );
      for( int j = 1; j < pPolygon->size(); j++ )
      {
         double lon1 = ( *pPolygon )[j - 1].lon;
         double lat1 = ( *pPolygon )[j - 1].lat;
         double lon2 = ( *pPolygon )[j].lon;
         double lat2 = ( *pPolygon )[j].lat;

         double ew1;
         double ns1;
         double ew2;
         double ns2;

         projPnt.getDistance( lon1, lat1, ew1, ns1 );
         projPnt.getDistance( lon2, lat2, ew2, ns2 );

         int x1 = ( ( ew1 + ( width / 2.0 ) ) / metersPerCell ) + 0.5;
         int y1 = ( ( -ns1 + ( width / 2.0 ) ) / metersPerCell ) + 0.5;
         int x2 = ( ( ew2 + ( width / 2.0 ) ) / metersPerCell ) + 0.5;
         int y2 = ( ( -ns2 + ( width / 2.0 ) ) / metersPerCell ) + 0.5;

         painter.drawLine( x1, y1, x2, y2 );
      }
   }
}