/*----------------------------------------------------------------------------*/
/*!
Copyright (c) 2018, 2019 by Christian Nowak a.k.a. Creestee a.k.a. chn
Contact: chnowak@web.de or creestee@creestee.com
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
3. This application may be used and redistributed in source code and binary
form free of charge for personal and for educational purposes.
4. Redistribution and usage of this application in source code or binary form
for commercial purposes requires explicit written permission from the
author.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of the RadarVolumeExplorer project.
*/
/*----------------------------------------------------------------------------*/

#define _USE_MATH_DEFINES
#include <math.h>

#include <QImage>
#include <QGLFunctions>

#include "RVEIsoModel.h"
#include <gl/GLU.h>

#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE  0x809D
#endif

/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveIsoModel::RveIsoModel() :
   RveModel( 0 ),
   m_Parameters( RveIsoParameters() ),
   m_Enabled( false )
{
   initializeGLFunctions();

   m_pCartesianModel = new RveCartesianModel( true );
}


/*---------------------------------------------------------------------------*/
/*!*/
/*---------------------------------------------------------------------------*/
RveIsoModel::~RveIsoModel()
{
   if( m_pCartesianModel )
   {
      delete m_pCartesianModel;
   }
}


void RveIsoModel::setParameters( RveIsoParameters parameters )
{
   if( m_Parameters != parameters )
   {
      m_Parameters = parameters;
      m_pCartesianModel->setParameters( parameters );
      if( isEnabled() )
         buildModel();
   }
}


RveIsoParameters RveIsoModel::parameters() const
{
   return( m_Parameters );
}


void RveIsoModel::setRawdata( RveRawdata *pRawdata )
{
   if( m_pRawdata != pRawdata )
   {
      m_pRawdata = pRawdata;
      m_pCartesianModel->setRawdata( pRawdata );
      if( isEnabled() )
         buildModel();
   }
}


void RveIsoModel::enable()
{
   m_pCartesianModel->enable();

   if( isEnabled() )
      return;

   m_Enabled = true;
   buildModel();
}


void RveIsoModel::disable()
{
   m_pCartesianModel->disable();

   if( !isEnabled() )
      return;

   m_Enabled = false;
   clearModel();

}


bool RveIsoModel::isEnabled() const
{
   return( m_Enabled );
}


void RveIsoModel::clearModel( double isolevel )
{
   clearVertexModel( isolevel );
   clearGLModel( isolevel );
}


void RveIsoModel::clearModel()
{
   QVector<double> levels = isoLevels();
   foreach( double isolevel, levels )
   {
      clearModel( isolevel );
   }
}


void RveIsoModel::paintGL( double isolevel )
{
   if( !m_GLModels.contains( isolevel ) ||
       !m_GLModelsIndices.contains( isolevel ) )
       return;

   if( !m_Parameters.hasIsoLevel( isolevel ) )
      return;

   glDisable( GL_LIGHTING );
   glDisable( GL_LIGHT0 );
   glDisable( GL_CULL_FACE );
   glEnable( GL_COLOR_MATERIAL );
   GLenum mode = GL_TRIANGLES;

   if(m_Parameters.isoParameters(isolevel).isoStyle == RveIsoParameters::Wireframe )
   {
      mode = GL_LINES;
   } else
   if(m_Parameters.isoParameters(isolevel).isoStyle == RveIsoParameters::Flat )
   {
      mode = GL_TRIANGLES;
   } else
   if(m_Parameters.isoParameters(isolevel).isoStyle == RveIsoParameters::Lighted )
   {
      mode = GL_TRIANGLES;
      glEnable( GL_LIGHTING );
      glEnable( GL_LIGHT0 );
      glEnable( GL_CULL_FACE );
      glEnable( GL_DEPTH_TEST );
   }

   glColor4f( m_Parameters.isoParameters( isolevel ).color.r,
	          m_Parameters.isoParameters( isolevel ).color.g,
	          m_Parameters.isoParameters( isolevel ).color.b,
              ( (double)m_Parameters.isoParameters( isolevel ).opacity ) / 100.0 );

   if( m_Parameters.isoParameters( isolevel ).opacity < 100 )
   {
      glDisable( GL_CULL_FACE );
//      glDisable( GL_DEPTH_TEST );
   } else
   {
      glEnable( GL_CULL_FACE );
      glEnable( GL_DEPTH_TEST );
   }

   glBindBuffer( GL_ARRAY_BUFFER, m_GLModels[isolevel] );
   glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, m_GLModelsIndices[isolevel] );
   glEnableClientState( GL_VERTEX_ARRAY );
//   glEnableClientState( GL_COLOR_ARRAY );
   glEnableClientState( GL_NORMAL_ARRAY );

   glVertexPointer( 3, GL_FLOAT, sizeof( Vertex ), (char *)NULL + (0) );
//   glColorPointer( 4, GL_FLOAT, sizeof( Vertex ), (char *)NULL + ( sizeof( XYZ ) ) );
   glNormalPointer( GL_FLOAT, sizeof( Vertex ), (char *)NULL + ( sizeof( XYZ ) + sizeof( RGBA ) ) );

   glDrawElements( mode, m_ModelsIndices[isolevel].size(), GL_UNSIGNED_INT, (char *)NULL + 0 );

   glDisableClientState( GL_VERTEX_ARRAY );
//   glDisableClientState( GL_COLOR_ARRAY );
   glDisableClientState( GL_NORMAL_ARRAY );
}

void RveIsoModel::paintGL()
{
   QVector<double> levels = isoLevels();
   qSort( levels );

   for( int i = levels.size() - 1; i >= 0; i-- )
   {
      paintGL( levels[i] );
   }
}


void RveIsoModel::clearGLModel( double isolevel )
{
   if( m_GLModels.contains( isolevel ) )
   {
      glDeleteBuffers( 1, &m_GLModels[isolevel] );
      m_GLModels.remove( isolevel );
   }

   if( m_GLModelsIndices.contains( isolevel ) )
   {
      glDeleteBuffers( 1, &m_GLModelsIndices[isolevel] );
      m_GLModelsIndices.remove( isolevel );
   }
}

void RveIsoModel::clearGLModel()
{
   QVector<double> isolevels = isoLevels();
   foreach( double isolevel, isolevels )
   {
      clearGLModel( isolevel );
   }
}


QVector<double> RveIsoModel::isoLevels()
{
   return( m_Parameters.isoLevels() );
}


void RveIsoModel::buildGLModel( double isolevel )
{
   glGenBuffers( 1, &m_GLModelsIndices[isolevel] );
   glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, m_GLModelsIndices[isolevel] );
   glBufferData( GL_ELEMENT_ARRAY_BUFFER,
                 m_ModelsIndices[isolevel].size() * sizeof( int ),
                 m_ModelsIndices[isolevel].data(),
                 GL_STATIC_DRAW );
   glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );

   glGenBuffers( 1, &m_GLModels[isolevel] );
   glBindBuffer( GL_ARRAY_BUFFER, m_GLModels[isolevel] );
   glBufferData( GL_ARRAY_BUFFER,
                 m_Models[isolevel].size() * sizeof( Vertex ),
                 m_Models[isolevel].data(),
                 GL_STATIC_DRAW );
   glBindBuffer( GL_ARRAY_BUFFER, 0 );
}


void RveIsoModel::buildVertexModel( double isolevel, RveModel::RGBA color )
{
   clearVertexModel( isolevel );

   polygonise( isolevel, color );

   for( int i = 0; i < m_Models[isolevel].size(); i++ )
   {
      m_ModelsIndices[isolevel].append( i );
   }
}


void RveIsoModel::clearVertexModel( double isolevel )
{
   if( m_Models.contains( isolevel ) )
      m_Models.remove( isolevel );
   if( m_ModelsIndices.contains( isolevel ) )
      m_ModelsIndices.remove( isolevel );
}


void RveIsoModel::clearVertexModel()
{
   QVector<double> isolevels = isoLevels();
   foreach( double isolevel, isolevels )
   {
      clearVertexModel( isolevel );
   }
}

void RveIsoModel::buildModel( double isolevel )
{
   clearModel( isolevel );
   RveModel::RGBA color;
   color.r = 1.0;
   color.g = 1.0;
   color.b = 1.0;
   color.a = 1.0;
   buildVertexModel( isolevel, color );
   buildGLModel( isolevel );
}



void RveIsoModel::buildModel()
{
   clearModel();
   QVector<double> levels = isoLevels();

   foreach( double isoLevel, levels )
   {
      buildModel( isoLevel );
   }
}


void RveIsoModel::polygonise( double isolevel, RveModel::RGBA color )
{
   for( int z = 0; z < m_pCartesianModel->parameters().layers() - 1; z++ )
   {
      for( int y = 0; y < m_pCartesianModel->parameters().cellsPerLayer() - 1; y++ )
      {
         for( int x = 0; x < m_pCartesianModel->parameters().cellsPerLayer() - 1; x++ )
         {
            polygonise( isolevel, color, x, y, z );
         }
      }
   }
}


void RveIsoModel::polygonise(
   double isolevel,
   RveModel::RGBA color,
   int x, int y, int z )
{
   GRIDCELL grid;
   TRIANGLE triangles[5];

   // Get the vertices from the cartesian model
   QVector<double> &data = m_pCartesianModel->data();
   QVector<Vertex> &vertices = m_pCartesianModel->vertices();
   int width = m_pCartesianModel->parameters().cellsPerLayer();
   int height = m_pCartesianModel->parameters().layers();
   for( int i = 0; i < 8; i++ )
   {
      int xofs = ((i + 1) >> 1 ) & 1;
      int yofs = (i & 4) >> 2;
      int zofs = (i & 2) >> 1;

      grid.p[i].x = vertices[(width * width * (z + zofs)) + (width * (y + yofs)) + (x + xofs)].point.x;
      grid.p[i].y = vertices[(width * width * (z + zofs)) + (width * (y + yofs)) + (x + xofs)].point.y;
      grid.p[i].z = vertices[(width * width * (z + zofs)) + (width * (y + yofs)) + (x + xofs)].point.z;
      grid.val[i] = data[(width * width * (z + zofs)) + (width * (y + yofs)) + (x + xofs)];
   }

   int numTriangles = Polygonise( grid, isolevel, triangles );

   for( int i = 0; i < numTriangles; i++ )
   {
      for( int j = 0; j < 3; j++ )
      {
         Vertex v;
         v.point.x = triangles[i].p[2 - j].x;
         v.point.y = triangles[i].p[2 - j].y;
         v.point.z = triangles[i].p[2 - j].z;
         v.color = color;
         m_Models[isolevel].append( v );
      }

      Vertex &p1 = m_Models[isolevel][m_Models[isolevel].size() - 3];
      Vertex &p2 = m_Models[isolevel][m_Models[isolevel].size() - 2];
      Vertex &p3 = m_Models[isolevel][m_Models[isolevel].size() - 1];
      Vertex v;
      v.point.x = p2.point.x - p1.point.x;
      v.point.y = p2.point.y - p1.point.y;
      v.point.z = p2.point.z - p1.point.z;
      Vertex w;
      w.point.x = p3.point.x - p1.point.x;
      w.point.y = p3.point.y - p1.point.y;
      w.point.z = p3.point.z - p1.point.z;

      double nx = ( (v.point.y * w.point.z) - (v.point.z * w.point.y) );
      double ny = ( (v.point.z * w.point.x) - (v.point.x * w.point.z) );
      double nz = ( (v.point.x * w.point.y) - (v.point.y * w.point.x) );
      double a = sqrt( nx * nx + ny * ny + nz * nz );
      nx /= a;
      ny /= a;
      nz /= a;
      p1.normal.x =
         p2.normal.x =
         p3.normal.x = nx;
      p1.normal.y = nx;
         p2.normal.y =
         p3.normal.y = ny;
      p1.normal.z =
         p2.normal.z =
         p3.normal.z = nz;
   }
}
